;;
;; todo: -GET-TERRACE-FILE needs to be sped up w/ some sort of caching
;;       -break the source into multiple files
;;       -all the tickets on redmine
;;
;; OVERVIEW
;;
;; Simply put: `Cl-Terrace' is the VC part of a hypothetical MVC framework written in
;; Common Lisp. It uses `Hunchentoot' for a webserver, `Djula' for HTML templating, and
;; a few abstractions over the url-as-file-in-the-filesystem metaphor for deciding how
;; to reply to HTTP requests
;;
;; TERRACE PROJECTS
;;
;; the View and Control of a simple website programmed using `cl-terrace' is referred to
;; as a ``terrace project''
;;
;; terrace projects are represented within a lisp image as symbols. the symbol becomes a
;; terrace project once it has been linked to a "site/" folder using the DEF-TERRACE
;; function
;;
;; a "sight/" folder must contain at least 1 of the following 3 subfolders:
;;
;;    1. "static/" -- the "static/" folder holds static content used by the project
;;    2. "template/" -- the "template/" folder holds Djula templates [and dictionaries]
;;                      used by the project
;;    3. "terrace/" -- the "terrace/" folder holds the project's ``terrace files''. the
;;                     contents of the "terrace/" directory control how the project
;;                     dispatches and handles individual HTTP requests. see
;;                     TERRACE FILES
;;
;; [note: running DEF-TERRACE hijacks the SYMBOL-FUNCTION slot of the symbol to
;; define a Hunchentoot dispatcher with the same name. so don't define a normal lisp
;; function with the same name afterwords]
;;
;; once you have turned a symbol into a terrace project you can to do any of the
;; following 3 things with it:
;;
;;    1. start a view server with it using the !START-TERRACE-SERVER/VIEW-ONLY function
;;      
;;       you can use a view server to preview interactive mock-ups of your website
;;       [see Djula's {% devel-dictionary %}, {% devel-value %}, and {% include %} tags
;;       for ideas].
;;
;;       After starting the server go to the root url (/) in your browser and it will
;;       guide you from there.
;;
;;    2. publish the project with the function !START-TERRACE-SERVER
;;
;;       once this is done your site is up and running and viewable in the browser!
;;
;;    3. push the project to HUNCHENTOOT:*DISPATCH-TABLE* and start Hunchentoot yourself
;;
;;       publishing the project manually makes it easier for multiple terrace projects
;;       to coexist behind the same Hunchentoot instance. Or for terrace projects to
;;       coexist with other lisp code replying to HTTP requests behind the same
;;       Hunchentoot image
;;
;;       note: remember you need to sync the project with the SYNC-TERRACE function
;;       before the dispatcher will do anything
;;
;; SYNCING PROJECTS
;;
;; calling the function SYNC-TERRACE on a project makes its Hunchentoot dispatcher aware
;; of any new changes or additions to the "terrace/" or "template/" directories. this
;; has to be done at least once before the project's Hunchentoot dispatcher will do
;; anything. this is done automatically by !START-TERRACE-SERVER
;;
;; SYNC-TERRACE is always aware of any new additions or changes to the terrace folder.
;; if you _delete_ a file, however, you have to call
;; RESET-TERRACE-FILE-FUNCTION-WRITE-DATE-CACHE on the project for SYNC-TERRACE to
;; notice [until someone gets around to fixing this]
;;
;; there is currently also  a bug in Djula that prevents templates from seeing changes
;; in their parent templates [templates that they extend using the {% extends %} and
;; {% block %} tags].
;;
;; to make a template aware of changes in its parents, call 
;; RESET-TERRACE-FILE-FUNCTION-WRITE-DATE-CACHE on the project before syncing it.
;; [fixing this problem has not been a high priority due to the easy workaround. it
;;  should be fixed eventually]
;;
;; in fact, RESET-TERRACE-FILE-FUNCTION-WRITE-DATE-CACHE can probably be considered
;; the equivalent of restarting Windows for terrace projects ;-)
;;
;; TERRACE FILES
;;
;; the "terrace/" folder contains the project's terrace files. terrace files control
;; how the project handles a particular set of possible HTTP requests. the positions of
;; terrace files within the "terrace/" folder [and how they relate to the url path of
;; the HTTP request] dictates which terrace  file will be responsible for replying to a
;; request
;;
;; terrace files are like normal lisp source files except their last form should return
;; the body of the reply as a string or array of bytes [like a Hunchentoot dispatcher],
;; or NIL, in which case the server will return a 404. 
;;
;; in simple cases this works something like PHP:
;;
;; this url:
;;
;;    http://localhost:8282/hello-world
;;
;; matches this file:
;;
;;    /hello-world.lisp
;;
;; [note: the root path in these examples is assumed to start at the "terrace/" folder]
;;
;; [".lisp" and ".cl" extensions are trimmed from the filename during this process]
;;
;; there is, however, much more you can do with terrace files. for instance if you have
;; the following file
;;
;;    /user/v.username.lisp
;;
;; containing the following lisp code:
;;
;;       (in-package :cl-terrace-user)
;;       (format nil "the username is ~A" (v "username"))
;;
;; then the following url
;;
;;    http://localhost:8282/user/bob
;;
;; with show up in the browser as
;;
;;    "the username is bob"
;;
;; see TERRACE FILE MATCHING RULES for more info on terrace files
;;
;; FUNCTIONS FOR USE INSIDE TERRACE FILES
;;
;; the following functions are exported from the CL-TERRACE package specifically for use
;; within terrace files
;;
;; RENDER (template-path &rest kwd-args &key &allow-other-keys)
;;
;;    -- renders and returns the Djula template pointed by `TEMPLATE-PATH'. the root
;;       path (/) for `TEMPLATE-PATH' is the "template/" folder
;;
;;       `KWD-ARGS' is a plist of keyword arguments that maps to variables used within
;;       the template
;;
;;       note: calling RENDER has a side effect: setting Hunchentoot's request reply
;;       content type to
;;
;;          "text/html; charset=utf-8"
;;
;; STATIC (static-path &optional content-type)
;;
;;    -- handles the file pointed to by `STATIC-PATH'. the root path (/) for
;;       `STATIC-PATH' is the "static/" folder
;;
;; FUNCALL-FILE (terrace-path &rest v-plist)
;;
;;    -- calls the terrace file pointed to by `TERRACE-PATH' [returning its value as if
;;       it were a normal lisp function]. the root path (/) for `TERRACE-PATH' is the
;;       "terrace/" folder
;;
;;       `V-PLIST' should be a plist of keys/values that will be visible to the V
;;        function within the body of the terrace file pointed to by `TERRACE-PATH'
;;
;; DICTIONARY-VALUE (variable template-path
;;                   &key (language djula:*language*) replace-regexps)
;;
;;    -- returns the value of the variable `VARIABLE' in the language `LANGUAGE' in
;;       the dictionary file pointed to by `TEMPLATE-PATH'
;;
;;       also returns a second value that is non-NULL if the value of `VARIABLE' in the
;;       language `LANGUAGE' is actually found [like GETHASH's second return value]
;;
;;       `REPLACE-REGEXPS' should be an alist keyed with regexps to be replaced by the
;;        values
;;
;;          (dictionary-value :variable "/my-dictionary.lisp")
;;
;;        -> "this variable has a {{nested-variable}}", t
;;
;;        (dictionary-value :variable
;;                          "/my-dictionary.lisp"
;;                          :replace-regexps '(("{{nested-variable}}" . "goose")))
;;
;;        -> "this variable has a goose", t
;;
;; V (key)
;;
;;    -- returns the value associated with the string `KEY'. V will look for the key in
;;       the following places:
;;
;;        1. the request url and how it relates to any "v." files or folders
;;           encountered while dispatching the request [see TERRACE FILE MATCHING RULES]
;;        2. any "special." files that are ancestors or siblings of the top-level
;;           terrace file responding to the request [see TERRACE FILE MATCHING RULES]
;;        3. the `v-plist' argument to FUNCALL-FILE [assuming there has been a call to
;;           FUNCALL-FILE somewhere up the stack]
;;
;; G (key)
;;
;;     -- returns the HTTP GET parameter named `key'.
;;
;;        G behaves differently from HUNCHENTOOT:GET-PARAMETER in that a GET parameter
;;        that has been supplied but has no value is represented as NIL [instead of an
;;        empty string]. there is a second return value that indicates whether the HTTP
;;        GET parameter was supplied or not
;;
;; P (key)
;;
;;     -- like G, but for POST parameters instead of GET parameters
;;
;; RETURNING FROM TERRACE FILES
;;
;; You can escape from a terrace file at any time by calling RETURN-FROM with the block
;; TERRACE-FILE. eg:
;;
;;    (return-from terrace-file "my webpage")
;;
;; note: there is no way if a terrace file can know if it's being called directly by 
;; Hunchentoot to handle a request or by FUNCALL-FILE from within another terrace file.
;; if you want to simulate returning from the _toplevel_ terrace file [the one handling
;; the request] throw a HUNCHENTOOT:HANDLER-DONE. eg:
;;
;;    (throw 'hunchentoot:handler-done "my webpage")
;;
;; TERRACE FILES AND COMMON LISP PACKAGES
;;
;; the default package of all terrace files is the :CL-USER package. use top-level
;; IN-PACKAGE forms to change *PACKAGE* while the file is being read [like you would do
;; with a normal lisp source file]
;;
;; `cl-terrace' comes with a :CL-TERRACE-USER package that uses the following packages:
;;    :COMMON-LISP
;;    :CL-TERRACE
;;    :DJULA
;;    :HUNCHENTOOT
;;    :LOGV
;;
;; UTF-8 ENCODING
;;
;; terrace files and Djula templates [and dictionaries] are always assumed to be encoded
;; in UTF-8. calling RENDER has the side effect of setting Hunchentoot's request reply
;; content type to
;;
;;    "text/html; charset=utf-8"
;;
;; TERRACE FILE MATCHING RULES
;;
;; URL: /foo/bar/baz
;; FILE: /foo/bar/baz.lisp
;;
;; -> serves result of evaluating the lisp code in "/foo/bar/baz.lisp"
;; [note .cl files work just the same as .lisp]
;;
;; URL: /foo/bar
;; FILE: /foo/bar.i/i.baz.lisp
;;
;; -> serves result of running the lisp code in "/foo/bar.i/i.baz.lisp"
;;
;; "i.baz.lisp" is the ``index-file'' of the folder "bar.i". Any folder that has an
;; index-file should have the suffix ".i". the obvious exception is the root
;; "terrace/" folder
;;
;; URL: /foo/bar/anything
;; FILE: /foo/bar/v.baz.lisp
;;
;; -> serves result of running the lisp code in "/foo/bar/v.baz.lisp"
;;
;; Within this code:
;;
;;       (v "baz") -> "anything"
;;
;; URL: /more/other/stuff
;; FILE: /v.foo/v.bar/v.baz.lisp
;;
;; -> serves result of running the lisp code in "/v.foo/v.bar/v.vaz.lisp"
;;
;; Within this code:
;;
;;    (v "foo") -> "more"
;;    (v "bar") -> "other"
;;    (v "baz") -> "stuff"
;;
;; URL: /foo/bar
;; FILE: /foo/bar.lisp
;; ALSO SEEN: /special.a.lisp -- contains the code (+ 1 2)
;;            /foo/special.b.lisp -- contains the code: (1+ (v "a"))
;;
;; -> serves result of running the lisp code in "/foo/bar.lisp".
;;
;; Within this code:
;;
;;    (v "a") -> 3
;;    (v "b") -> 4
;;

(defpackage :cl-terrace
  (:use :cl :f-underscore :logv)
  (:export ; defining-projects
           :def-terrace

	   ; functions for use within terrace files
	   :render
	   :static
	   :funcall-file
	   :dictionary-value
	   :v
	   :g
	   :p

	   ; block you can RETURN-FROM to escape from terrace files
	   :terrace-file

	   ; syncing projects
	   :sync-terrace
	   :clear-terrace-file-function-write-date-cache

	   ; publishing projects
           :*terrace-server*
	   :!start-terrace-server
	   :!stop-terrace-server

	   ; view server
	   :!start-terrace-server/view-only

	   ; stuff that is probably useful for debugging
	   :*terrace-project*
	   :*terrace-path*
	   :*v-plist*
	   :get-terrace-file
	   :special-path-alist->plist
	   :funcall-path))

(defpackage :cl-terrace-user
  (:use :cl :cl-terrace :hunchentoot :djula :logv))

(in-package :cl-terrace)

; configure these

(defvar *boring-extentions* (list "lisp" "cl")
  "These file extentions are ignored by `cl-terrace' when mapping between a request
url-path and the filesystem. So the url-path

   /foo/bar

can find the file

   /foo/bar.lisp

if the string \"lisp\" is found in *BORING-EXTENTIONS*
")

; special

(defvar *terrace-project*)

(defvar *terrace-path*)

(defvar *terrace-server* nil)

(defvar *v-plist*)

; templates used internally

(defparameter *source-folder* (asdf:component-pathname (asdf:find-component (asdf:find-system :cl-terrace)
									    "cl-terrace")))

(defparameter *internal-template-folder*
  (merge-pathnames "templates/" *source-folder*))

(defparameter *view-only-index-page-template*
  (merge-pathnames "view-only-index-page.html" *internal-template-folder*))

(defparameter *view-only-show-dictionary-template*
  (merge-pathnames "view-only-show-dictionary.html" *internal-template-folder*))

; util

(defmacro aif (test &optional (then 'it) else)
  `(let ((it ,test))
     (if it
	 ,then
	 ,else)))

(defun .getf (place indicator)
  "like GETF but returns a second value indicating the lookup's success or failure
[like GETHASH's PRESENT-P return value]"
  (when place
    (destructuring-bind (k v . rest) place
      (if (eql k indicator)
	  (values v place)
	  (.getf rest indicator)))))

(defun .getf-equal (place indicator)
  "like GETF but uses EQUAL instead of EQL and returns a second value indicating the
lookup's success or failure [like GETHASH's PRESENT-P return value]"
  (when place
    (destructuring-bind (k v . rest) place
      (if (equal k indicator)
	  (values v place)
	  (.getf-equal rest indicator)))))

(defun .matches-prefix (prefix string)
  "
   (.matches-prefix \"foo-\" \"foo-bar\")

   -> \"bar\"

   (.matches-prefix \"foo-\" \"baz-bar\")

   -> NIL
"
  (let ((prefix-length (length prefix)))
    (when (>= (length string) prefix-length)
      (let ((m (mismatch prefix string :test 'char=)))
	(if (or (null m)
		(= m prefix-length))
	    (subseq string m))))))

(defun .in-package-form-p (form)
  "if `FORM' is a list that looks like

   (in-package <package-designator>)

then returns <package-designator> otherwise returns NIL
"
  (and (listp form)
       (null (cddr form))
       (eql (first form) 'in-package)
       (second form)))

(defun .read-from-string-understanding-in-package-forms (string)
  "like READ-FROM-STRING but returns a list of all forms in the string, not just the
first form. understands toplevel IN-PACKAGE forms while reading, but discards them after
they are read"
  (let (forms
	(done (gensym "done"))
	(*package* (find-package :cl-user)))
    (with-input-from-string (in string)
      (do ((form #1=(read in nil done) #1#))
	  ((eql form done) (nreverse forms))
	(if (.in-package-form-p form)
	    (setq *package* (find-package (.in-package-form-p form)))
	    (push form forms))))))

(defun .trim-boring-extentions (filename)
  "removes the extension of `FILENAME' if is found in *BORING-EXTENTIONS*"
  (if (find (pathname-type filename) *boring-extentions* :test 'equal)
      (pathname-name filename)
      filename))

; terrace/static/template paths

(defun .get-path (local-name folder-key)
  "returns the file or folder pointed to by the pathname designator `P', where the root
path / is seen as

   (GET *TERRACE-PROJECT* FOLDER-KEY)
"
  (merge-pathnames (string-trim '(#\/) (namestring local-name))
		   (get *terrace-project* folder-key)))

(defun .try-to-check-path (p folder-key)
  "warns to LOGV:*LOG-OUTPUT* if it can figure out that `P' points to a file or folder
that doesn't exist in

   (get *TERRACE-PROJECT* FOLDER-KEY)
"
  (if (and (or (stringp p)
	       (pathnamep p))
	   (symbolp folder-key)
	   (boundp '*terrace-project*)
	   (boundp '*terrace-path*)
	   (not (aif (.get-path p folder-key)
		     (if (cl-fad:directory-pathname-p p)
			 (cl-fad:directory-exists-p it)
			 (cl-fad:file-exists-p it)))))
      (logv:format-log "<<<~A ~A doesn't exist: ~S>>>"
		       folder-key
		       (if (cl-fad:directory-pathname-p p)
			   "folder"
			   "file")
		       p))
  (values))
  
; identifying files and folders

(defun .identify-filename (name)
  "returns two values:

   1. a keyword identifying a special Terrace prefix or NIL if `NAME' does not have one
   2. the token created by the match

eg

   (.identify-filename \"v.foo.cl\")

   -> :v, \"foo.cl\"

   (.identify-filename \"foo.cl\")

   -> NIL, \"foo.cl\"
"
  (flet ((% (<id> <token>)
	   (if <token>
	       (return-from .identify-filename (values <id> <token>)))))
    (let ((name (namestring name)))

      ; v.foo.cl -> :v, "foo.cl"
      (% :v (.matches-prefix "v." name))

      ; special.foo.cl -> :special, "foo.cl"
      (% :special (.matches-prefix "special." name))

      ; i.foo.cl -> :i, "foo.cl"
      (% :i (.matches-prefix "i." name))

      ; foo.cl -> NIL, "foo.cl"
      (values nil name))))

(defun .identify-foldername (name)
  "returns two values:

   1. a keyword identifying a special Terrace prefix/suffix or NIL if `NAME' does not
      have any special prefix/suffix
   2. the token created by the match

eg

   (.identify-foldername \"v.foo\")

   -> :v, \"foo\"

   (.identify-foldername \"foo.i\")

   -> :i, \"foo\"

   (.identify-filename \"foo\")

   -> NIL, \"foo\"
"
  (flet ((% (<id> <token>)
	   (if <token>
	       (return-from .identify-foldername (values <id> <token>)))))
    (let ((name (namestring name)))

      ; v.foo -> :v, "foo"
      (% :v (.matches-prefix "v." name))

      ; foo.i -> :i, "foo"
      (% :i (if (equal (pathname-type name) "i")
		(pathname-name name)))

      ; foo -> NIL, "foo"
      (values nil name))))

(defmacro .with-identified-file ((path identify &optional token) &body body)
  "if `PATH' can be identified as an `IDENTIFY' type of file then executes `BODY'.
Otherwise returns NIL.

If `TOKEN' is supplied, then it is bound to the token resulting from the identification
within the scope of `BODY'

   (.with-identified-file (\"/foo/v.bar.cl\" :v <token>)
     (values t <token>))

   -> t, \"bar.cl\"

   (.with-identified-file (\"/foo/v.bar.cl\" :i <token>)
     (values t <token>))

   -> NIL
"
  (let ((<type> (gensym "type")))
    `(multiple-value-bind (,<type> ,@(if token
					 `(,token)))
	 (.identify-filename (djula-util:filename ,path))
       (case ,<type>
	 ((,identify) ,@body)
	 (otherwise nil)))))

(defmacro .with-identified-folder ((path identify &optional token) &body body)
  "if `PATH' can be identified as an `IDENTIFY' type of folder then executes `BODY'.
Otherwise returns NIL.

If `TOKEN' is supplied, then it is bound to the token resulting from the identification
within the scope of `BODY'

   (.with-identified-folder (\"/foo/v.bar/\" :v <token>)
     (values t <token>))

   -> t, \"bar\"

   (.with-identified-folder (\"foo/v.bar/\" :i <token>)
     (values t <token>))

   -> NIL
"
  (let ((<type> (gensym "type")))
    `(multiple-value-bind (,<type> ,@(if token
					 `(,token)))
	 (.identify-foldername (djula-util:foldername ,path))
       (case ,<type>
	 ((,identify) ,@body)
	 (otherwise nil)))))

; matching terrace files from urls

(defun .get-normal-match/file (folder name)
  "returns any files in `FOLDER' that match the name `NAME' [does not return any \"v.\"
file matches]"
  (some (f_ (.with-identified-file (_ nil)
	      (if (equal (.trim-boring-extentions (djula-util:filename _)) name)
		  _)))
	(djula-util:non-boring-files folder)))

(defun .get-v-match/file (folder)
  "returns the first \"v.\" file found in `FOLDER'"
  (mapc (f_ (.with-identified-file (_ :v <token>)
	      (return-from .get-v-match/file (values _ <token>))))
	(djula-util:non-boring-files folder))
  nil)

(defun .get-normal-match/folder (folder name)
    "returns any folders in `FOLDER' that match the name `NAME' [does not return any
\"v.\" folder matches]"
  (some (f_ (block <f_>
	      (.with-identified-folder (_ nil)
		(if (equal (djula-util:foldername _) name)
		    (return-from <f_> _)))
	      (.with-identified-folder (_ :i <token>)
		(if (equal <token> name)
		    (return-from <f_> _)))))
	(djula-util:non-boring-folders folder)))

(defun .get-v-match/folder (folder)
  "returns the first \"v.\" folder found in `FOLDER'"
  (mapc (f_ (.with-identified-folder (_ :v <token>)
	      (return-from .get-v-match/folder (values _ <token>))))
	(djula-util:non-boring-folders folder))
  nil)

(defun .get-index-file (folder)
  "if `FOLDER' is has the suffix \".i\" or `ROOT-FOLDER-P' is T, then `FOLDER' is an
\".i\" folder. return the first file with the prefix \"i.\""
  (some (f_ (.with-identified-file (_ :i)
	      _))
	(djula-util:non-boring-files folder)))

(defun .get-index-match (folder name)
  (some (f_ (.with-identified-folder (_ :i <token>)
	      (when (equal <token> name)
		(.get-index-file _))))
	(djula-util:non-boring-folders folder)))

(defun .get-special-path-alist (folder)
  "returns an alist of

    (token . path)

key/vals made from any \"special.\" files seen in folder"
  (mapcan (f_ (.with-identified-file (_ :special <token>)
		(list (cons (.trim-boring-extentions <token>) _))))
	  (djula-util:non-boring-files folder)))

(defun .get-terrace-file (url-path-list path)
  "Tries to match `URL-PATH' with `PATH' [which may be a file or a folder]. If there is
no match returns NIL. If there is a match returns the 3 values documented in
GET-TERRACE-FILE"
  (if (null url-path-list)

      ;; `URL-PATH-LIST' is empty. This means that `PATH' is probably the root folder
      ;; and GET-TERRACE-FILE was called with a url "/". Look for an "i." file in `PATH'
      (if (cl-fad:directory-pathname-p path)
	  (aif (.get-index-file path)
	       (values it nil (.get-special-path-alist path))))

      ;; `URL-PATH-LIST' is not empty
      (destructuring-bind (1st . rest) url-path-list
	(if (null rest)

	    ;; `URL-PATH-LIST' points to a file

	    ;; maybe a "v." match
	    (multiple-value-bind (<v-match> <v-token>) (.get-v-match/file path)
	      (if <v-match>
		  (values <v-match>
			  (list (.trim-boring-extentions <v-token>) 1st)
			  (.get-special-path-alist path))
		  
		  
		  (values (or ;; maybe a normal match
			     (.get-normal-match/file path 1st)

			     ;; maybe a ".i" directory with an "i." file
			     (.get-index-match path 1st))

			  nil
			  (.get-special-path-alist path))))

	    ;; `URL-PATH-LIST' points to a directory
	    (progn

	      ;; maybe a "v." match
	      (multiple-value-bind (match <token>) (.get-v-match/folder path)
		(if match
		    (multiple-value-bind (resource v-url-plist special-v-path-alist)
			(.get-terrace-file rest match)
		      (if resource
			  (return-from .get-terrace-file
			               (values resource
					       (list* (.trim-boring-extentions <token>) 1st v-url-plist)
					       (append (.get-special-path-alist path)
						       special-v-path-alist)))))))

	      ;; maybe a normal match
	      (let ((match (.get-normal-match/folder path 1st)))
		(if match
		    (multiple-value-bind (resource v-url-plist special-v-path-alist)
			(.get-terrace-file rest match)
		      (if resource
			  (return-from .get-terrace-file
			               (values resource
					       v-url-plist
					       (append (.get-special-path-alist path)
						       special-v-path-alist))))))))))))


(defun get-terrace-file (url-path)
  "Tries to match `URL-PATH' with the \"terrace/\" folder of *TERRACE-PROJECT*. If there
is no match returns NIL. If there is a match returns 3 values:

   1. the pathname of the matching file
   2. the v-plist derived by combining `URL-PATH' with `PATH'
   3. the \"special.\" path alist seen while making the match
      [see SPECIAL-PATH-ALIST->PLIST]
"
  (.get-terrace-file (split-sequence:split-sequence #\/
						    url-path
						    :remove-empty-subseqs t)
		     (get *terrace-project* 'terrace-folder)))

; compile terrace / template files

(defun .log-function-compilation (path description recompiling-p)
  (logv:format-log  "~A~Afile: \"/~A\""
		    (if recompiling-p
			"Recompiling"
			"Compiling")
		    (if description
			(format nil " ~A " description)
			" ")
		    (enough-namestring path (get *terrace-project* 'site-folder))))

(defun .compile-terrace-file (path new-p)

  ;; first log what's happening
  (.log-function-compilation path (.identify-filename path) new-p)

  ;; then compile the function
  (let* ((*terrace-path* path))
    (compile nil
	     (coerce `(lambda ()
			(let ((*terrace-path* ,path))
			  (block terrace-file
			    ,@(.read-from-string-understanding-in-package-forms
			       (djula-util:slurp-utf-8-file path)))))
		     'function))))

;; we dont use DJULA:CREATE-DICTIONARY-CACHE so we can give better log output

(defun .compile-dictionary-file (path new-p)

  ;; first log whats happening
  (.log-function-compilation path "dictionary" new-p)

  ;; finally return a thunk that simply
  ;; returns the contents of the
  ;; dictionary
  (let ((d (djula::.read-dictionary path)))
    (constantly d)))

(defun .compile-template-file (path new-p)

  ;; first log what's happening
  (.log-function-compilation path "template" new-p)

  ;; finally compile the template file
  (djula:compile-template path
			  :template-folder (get *terrace-project* 'template-folder)))

(defun .funcall-file (path cache-name &rest args)
  (aif (djula-util:get-file-function path
				     (get *terrace-project* cache-name))
       (apply it args)))

; dispatching terrace files from urls

(defun special-path-alist->plist (alist)
  "takes a special-path-alist [like the one returned as the 3rd return value of
GET-TERRACE-FILE], which is filled with

   (token . path)

key/vals and returns a plist where the keys are the tokens from the alist and the vals
are the results of calling FUNCALL-FILE on the paths of the alist"
  (when alist
    (destructuring-bind ((<k> . path) . rest) alist
      (list* <k>
	     (.funcall-file path 'terrace-cache)
	     (special-path-alist->plist rest)))))

(defun .dispatch-url-path (url-path)
  "if there is a terrace file matching `URL-PATH' [*TERRACE-PROJECT* must be bound to a
Terrace project] then returns a handler that, when called, replies to the request.
Otherwise returns NIL"
  (multiple-value-bind (path url-v-plist special-path-alist) 
      (get-terrace-file url-path)
    (if path
	(let ((terrace-project *terrace-project*))
	  (f0 (let* ((*terrace-project* terrace-project)
		     (*v-plist* (nconc url-v-plist (special-path-alist->plist special-path-alist))))
		(aif (.funcall-file (namestring path) 'terrace-cache)
		     it
		     ;; if the terrace file returns NIL throw a 404
		     (progn
		       (setf (hunchentoot:return-code) hunchentoot:+http-not-found+)
		       (throw 'hunchentoot:handler-done nil)))))))))

(defun funcall-path (url-path)
  (aif (.dispatch-url-path url-path)
       (funcall it)))

; functions for use inside terrace files

(defun render (template-path &rest plist)
"sets the Hunchentoot reply content type to

   \"text/html; charset=utf-8\"

and renders the Djula template pointed to by `TEMPLATE-PATH' [where / is seen as the
\"template/\" directory of the current Terrace project]

`PLIST' should be a plist of template variables.
"
  ;; setup hunchentoot utf-8 stuff
  (setf (hunchentoot:reply-external-format) (flexi-streams:make-external-format :utf-8)
	(hunchentoot:content-type)          "text/html; charset=utf-8")
      ;;
      ;;         should NIL from funcall-file throw 404?
      ;;
      (apply '.funcall-file
	     (.get-path template-path 'template-folder)
	     'template-cache
	     plist))

(define-compiler-macro render (&whole form template-path &rest plist)
  (declare (ignore plist))
  (.try-to-check-path template-path 'template-folder)
  form)

(defun static (static-path &optional content-type)
"calls HUNCHENTOOT:HANDLE-STATIC-file on the file pointed to by `STATIC-PATH' [where /
is seen as the \"static/\" directory of the current Terrace project"
  (hunchentoot:handle-static-file (.get-path static-path 'static-folder)
				  content-type))

(define-compiler-macro static (&whole form static-path &optional content-type)
  (declare (ignore content-type))
  (.try-to-check-path static-path 'static-folder)
  form)

(defun funcall-file (terrace-path &rest v-plist)
  "calls the Terrace file pointed to by `TERRACE-PATH' [where / is seen as the
\"terrace/\" directory of the current Terrace project] as if it were a function and
returns it's value

 `V-PLIST' should be aplist of key/val pairs for the V function
"
  (let ((*v-plist* (append v-plist *v-plist*)))
    (.funcall-file (.get-path terrace-path 'terrace-folder)
		   'terrace-cache)))

(define-compiler-macro funcall-file (&whole form terrace-path &rest v-plist)
  (declare (ignore v-plist))
  (.try-to-check-path terrace-path 'terrace-folder)
  form)

(defun v (key)
  "finds the value associated with the key `KEY' by looking into the plist *V-PLIST*.
*V-PLIST* might be populated with key/val pairs from

   1. \"special.\" files that are ancestors or siblings to the top-level Terrace file
      responding to the request
   2. \"v.\" files or folders that have an ancestor-or-self relationship to the top-level
      Terrace file responding to the request
   3. the FUNCALL-FILE function [assuming the current Terrace file has been called by
      some parent terrace file]
"
  (.getf-equal *v-plist* key))

(defun g (key)
  "returns two values:

   1. the HTTP GET parameter named `KEY'. Returns NIL if it has not been supplied or
it has been supplied but has no value.
   2. whether or not the GET parameter was supplied
"
  (aif (hunchentoot:get-parameter key)
       (values (if (not (zerop (length it)))
		   it)
	       t)))

(defun p (key)
  "returns two values:

   1. the HTTP POST parameter named `KEY'. Returns NIL if it has not been supplied or
it has been supplied but has no value.
   2. whether or not the POST parameter was supplied
"
  (aif (hunchentoot:post-parameter key)
       (values (if (not (zerop (length it)))
		   it)
	       t)))

(defun dictionary-value (variable template-path &key (language djula:*language*) replace-regexps)
  "returns two values:

   1. the value of the variable `VARIABLE' in the dictionary `TEMPLATE-PATH' in the
      language `LANGUAGE'
   2. whether or not the variable exists in the dictionary

   (dictionary-value :variable \"/my-dictionary.lisp\")

   -> \"this variable has a {{nested-variable}}\", t

   (dictionary-value :variable
                      \"/my-dictionary.lisp\"
                     :replace-regexps '((\"{{nested-variable}}\" . \"goose\")))

   -> \"this variable has a goose\"

"
  (aif (assoc variable (.funcall-file (.get-path template-path
						 'template-folder)
				      'dictionary-cache))
       (multiple-value-bind (v present-p) (.getf (rest it) language)
	 (when present-p
	   (dolist (% replace-regexps v)
	     (destructuring-bind (old . new) %
	       (setf v (cl-ppcre:regex-replace-all old v new))))))))

(define-compiler-macro dictionary-value (&whole form variable template-path &key language replace-regexps)
  (declare (ignore variable language replace-regexps))
  (.try-to-check-path template-path 'template-folder)
  form)

; projects

; maybe store template-folder again as dictionary-folder?

(defun def-terrace (project-name
		    site-folder
		    &key (publish-static-p t))
"Creates a Terrace project named `PROJECT-NAME' with a \"site/\" folder `SITE-FOLDER'
[note: this creates a Hunchentoot dispatcher named `PROJECT-NAME', so you can't ever
define a function named `PROJECT-NAME' afterwords].

`SITE-FOLDER' should contain 3 subfolders:

   1. terrace/ -- holds Terrace files [\".lisp\" or \".cl\" files] that are used to
                  dispatch HTTP requests. Make sure to read about the rules for matching
                  url-paths to Terrace Files at the top of this source file. 

   2. template/ -- holds Djula templates. templates in this folder can be rendered with
                   the function TEMPLATE

   3. static/ -- holds static files used to dispatch an HTTP request

if `PUBLISH-STATIC-P' is non-NULL then the `PROJECT-NAME' Hunchentoot dispatcher
publishes the contents of the \"static/\" folder. Otherwise the \"static/\" folder is
ignored. [It's much faster to have some other webserver such as nginx or Apache sitting
in front of Hunchentoot serving your static files]
"
  (let* ((terrace-folder  (merge-pathnames "terrace/" site-folder))
	 (template-folder (merge-pathnames "template/" site-folder))
	 (static-folder   (merge-pathnames "static/" site-folder))
	 (static-dispatcher (if publish-static-p
				(hunchentoot:create-folder-dispatcher-and-handler "/" static-folder))))

    ;; create a "terrace/" folder dispatcher named "project-name
    (setf (symbol-function project-name)
	  (lambda (hunchentoot:*request*)
	    (let ((*terrace-project* project-name))

	      (or ;; maybe dispatch terrace file
	          (.dispatch-url-path (hunchentoot:script-name))

	          ;; maybe dispatch static file
		  (if static-dispatcher
		      (funcall static-dispatcher hunchentoot:*request*))))))

    ;; set symbol-plist
    (setf ;; save `SITE-FOLDER' and subfolders
          (get project-name 'site-folder)       site-folder
          (get project-name 'terrace-folder)    terrace-folder
          (get project-name 'template-folder)   template-folder
          (get project-name 'dictionary-folder) template-folder
          (get project-name 'static-folder)     static-folder

	  ;; save keyword options
	  (get project-name 'publish-static-p) publish-static-p

	  ;; create project caches
	  (get project-name 'terrace-cache)    (djula-util:create-file-function-cache '.compile-terrace-file)
	  (get project-name 'template-cache)   (djula-util:create-file-function-cache '.compile-template-file
											   :test (complement #'djula::.dictionary-p))
	  (get project-name 'dictionary-cache) (djula-util:create-file-function-cache '.compile-dictionary-file
											   :test 'djula::.dictionary-p)

	  ;; finalize
	  (get project-name 'terrace-project-p) t)

    ;; make `PROJECT-NAME' evaluate to itself
    (setf (symbol-value project-name)
	  project-name)

    project-name))

(defun sync-terrace (project &key (sync-template-folder t) (sync-terrace-folder t))
  "makes sure all the functions compiled from files in `PROJECT's \"terrace/\" and
\"template/\" folders are up to date, recompiling them if the're not"
  (if (not (get project 'terrace-project-p))
      (error "<<<~S does not name a terrace project [see DEF-TERRACE]>>>" project)
      (flet ((syncit (folder-name cache-name)
		 (aif (get project folder-name)
		      (if (cl-fad:directory-exists-p it)
			  (let ((*terrace-project* project))
			    (logv:format-log "Syncing ~A" folder-name)
			    (djula-util:sync-file-function-cache it
								 (get project cache-name)))
			  (logv:format-log "<<<There is no ~A folder in \"site/\">>>"
					   folder-name))
		      (logv:format-log "<<<~A does not have a ~A folder>>>"
				       project
				       folder-name))))
	(logv:format-log "\"site/\": ~A" (get project 'site-folder))
	(when sync-template-folder
	  (syncit 'template-folder 'template-cache)
	  (syncit 'dictionary-folder 'dictionary-cache))
	(when sync-terrace-folder
	  (syncit 'terrace-folder 'terrace-cache))))
  project)

(defun clear-terrace-file-function-write-date-cache (project)
  (if (not (get project 'terrace-project-p))
      (logv:format-log "<<<~S does not name a terrace project [see DEF-TERRACE]>>>"
		       project)
      (values

	;; clear template cache
	(aif (get project 'template-cache)
	     (djula-util:clear-file-function-write-date-cache it))

	;; clear dictionary cache
	(aif (get project 'dictionary-cache)
	     (djula-util:clear-file-function-write-date-cache it))

	;;clear terrace file cache
	(aif (get project 'terrace-cache)
	     (djula-util:clear-file-function-write-date-cache it)))))

(defun .start-terrace-server (project-name action-thunk &rest hunchentoot-kwds)
  "
   1. logs a bunch of stuff about the Terrace project PROJECT-NAME
   2. performs `ACTION-THUNK'
   3. starts a Hunchentoot server using `HUNCHENTOOT-KWDS' as keyword arguments, setting
      *TERRACE-SERVER* to the resultant hunchentoot server [so it can be stopped by
      !STOP-TERRACE-SERVER]
"

  ;; run all sorts of checks to make sure some things aren't seriously wrong
  (if *terrace-server*
      (error "A Terrace server is already running [see *TERRACE-SERVER*]. run !STOP-TERRACE-SERVER to stop it before trying to start it again")
      (if (not (get project-name 'terrace-project-p))
	  (error "The symbol ~S does not name Terrace project. You need to run DEF-TERRACE to define a Terrace project before you can start it]"
		 project-name)
	  (let ((site-folder (get project-name 'site-folder)))
	    (if (not (cl-fad:directory-exists-p site-folder))
		(error "Cannot start Terrace project ~S. There is no \"site/\" folder at ~A"
		       project-name
		       site-folder)

		;; nothing seems to be seriously wrong
		(progn

		  ;; log DEF-TERRACE keyword args
		  (let ((publish-static-p (get project-name 'devel-only-p)))
		    (logv publish-static-p))
		    
		  ;; log "template/", "terrace/", and "static/" folders
		  (flet ((log-folder (name)
			   (let ((folder (get project-name name)))
			     (if (and folder
				      (cl-fad:directory-exists-p folder))
				 (logv:format-log "There are ~A non-boring file~:*~p and ~A non-boring directories in the ~A"
						  (length (djula-util:non-boring-files folder))
						  (length (djula-util:non-boring-folders folder))
						  name)
				 (logv:format-log "<<<There is no ~A!!>>>"
						  name)))))
		    (log-folder 'static-folder)
		    (log-folder 'template-folder)
		    (log-folder 'terrace-folder))

		  ;; perform action
		  (funcall action-thunk)

		  ;; start-server
		  (logv:format-log "Starting Hunchentoot")
		  (setf *terrace-server* (apply 'hunchentoot:start-server hunchentoot-kwds))

		  ;; warn if LOGV:*LOG-OUTPUT* is t
		  (when (eql logv:*log-output* t)
		    (logv:format-log "-------------------------------------------------------------------------")
		    (logv:format-log "LOGV:*LOG-OUTPUT* is T, which means you won't be able to see stuff logged")
		    (logv:format-log "while replying to HTTP requests in the repl because Hunchentoot rebinds")
		    (logv:format-log "*STANDARD-OUTPUT*. Run (SETF LOGV:*LOG-OUTPUT* *STANDARD-OUTPUT*) to see")
		    (logv:format-log "logs in the current repl when replying to HTTP requests, or better yet,")
		    (logv:format-log "set LOGV:*LOG-OUTPUT* to the path of a log file")
		    (logv:format-log "-------------------------------------------------------------------------"))

		  *terrace-server*))))))

(defun !start-terrace-server (project-name &rest hunchentoot-kwds)
  "
   1. logs a bunch of stuff
   2. Syncs the \"site/\" folder of the project `PROJECT-NAME'
   2. starts a Hunchentoot server using `HUNCHENTOOT-KWDS' as keyword arguments
   3. publishes the Terrace project `PROJECT-NAME'.
"
  (apply '.start-terrace-server
	 project-name
	 (f0 ;; sync project
	     (sync-terrace project-name)

	     ;; update dispatch-table
	     (logv:format-log "Updating Hunchentoot Dispatch Table")
	     (setf hunchentoot:*dispatch-table*
		   (list* project-name (remove project-name hunchentoot:*dispatch-table*))))
	 hunchentoot-kwds))
	
(defun !stop-terrace-server ()
  "stops the Hunchentoot server created by !START-TERRACE-SERVER"
  (if *terrace-server*
      (progn
	(hunchentoot:stop-server *terrace-server*)
	(setf *terrace-server*             nil
	      hunchentoot:*dispatch-table* (list 'hunchentoot:default-dispatcher)))
      (error "There is no terrace server running [see *TERRACE-SERVER*]")))